#include <errno.h>
#include <sys/acl.h>

static int acl_storage_order[] = { 
  ACL_USER_OBJ, 
  ACL_MASK, 
  ACL_USER, 
  ACL_GROUP_OBJ, 
  ACL_GROUP, 
  ACL_OTHER_OBJ 
};

#include <stdio.h>
#define acl_debug(f, a...) fprintf(stderr, f, ##a)

static
int check_acl (struct linux_acl *lacl)
{
	int i = 0;
	int error = 0;
	int state = 0;

	if (lacl->acl_valid != LINUX_ACL_VALID) {
		acl_debug ("No valid flag for ACL\n");
		error = -EINVAL;
	}
	for (i = 0; i < lacl->acl_acle_count; i++) {
		if (lacl->acl_acle[i].acle_valid != LINUX_ACL_VALID) {
			acl_debug ("No valid flag at pos %d\n", i);
			error = -EINVAL;
		}
		switch (lacl->acl_acle[i].acle_type) {
		case ACL_USER_OBJ:
			if (state == 0) {
				state = 1;
			} else {
				acl_debug ("Unexpected ACL_USER_OBJ at pos %d\n", i);
				error = -EINVAL;
			}
			break;
		case ACL_MASK:
			if (state == 1) {
				state = 2;
			} else {
				acl_debug ("Unexpected ACL_MASK at pos %d\n", i);
				error = -EINVAL;
			}
			break;
		case ACL_USER:
			if (state != 2) {
				acl_debug ("Unexpected ACL_USER at pos %d\n", i);
				error = -EINVAL;
			}
			break;
		case ACL_GROUP_OBJ:
			if (state == 1) {
				state = 3;
			} else if (state == 2) {
				state = 4;
			} else {
				acl_debug ("Unexpected ACL_GROUP_OBJ at pos %d\n", i);
				error = -EINVAL;
			}
			break;
		case ACL_GROUP:
			if (state != 4) {
				acl_debug ("Unexpected ACL_GROUP at pos %d\n", i);
				error = -EINVAL;
			}
			break;
		case ACL_OTHER_OBJ:
			if (state != 3 && state != 4) {
				acl_debug ("Unexpected ACL_OTHER_OBJ at pos %d\n", i);
				error = -EINVAL;
			}
			state = 5;
			break;
		default:
			acl_debug ("Bad type %d at pos %d\n",
				   lacl->acl_acle[i].acle_type, i);
			error = -EINVAL;
			break;
		}
		if (lacl->acl_acle[i].acle_perms & ~7) {
			acl_debug ("Bad permission (%o) in ACL at pos %d\n",
				   lacl->acl_acle[i].acle_perms,
				   i);
			error = -EINVAL;
		}
	}
	if (state != 5) {
		acl_debug ("Missing a required entry\n");
		error = -EINVAL;
	}
	return error;
}

static
int		acl_set (const int set_op, const void * file,
			 const acl_type_t type_d, const acl_t acl_d)
{
	acl_entry_t		acle;
	struct linux_acl	*lacl;
	acl_tag_t		tag_type;
	int			i,j;
	int			error;

	if (acl_valid (acl_d, &acle) != 0)
		return -1;
	lacl = (struct linux_acl *)malloc (sizeof (struct linux_acl)
					   + (sizeof (struct linux_acl_entry)
					      * acl_d->acl_used_acle));
	if (!lacl) {
		errno = ENOMEM;
		return -1;
	}
	lacl->acl_valid = LINUX_ACL_VALID;
	lacl->acl_acle_count = acl_d->acl_used_acle;
	i = 0;
	for (j = 0; j < NO_ACL_TYPES; j++) {
		tag_type= acl_storage_order[j];
		for (acle = acl_d->acl_faclep; acle; acle = acle->acle_next)
			if (acle->acle_tag_type == tag_type) {
				lacl->acl_acle[i].acle_valid = LINUX_ACL_VALID;
				lacl->acl_acle[i].acle_type = acle->acle_tag_type;
				lacl->acl_acle[i].acle_perms = acle->acle_perms;
				lacl->acl_acle[i].acle_tag = acle->acle_tag;
				i++;
			}
	}
  check_acl(lacl);
	error = acl_ctl (set_op, file, type_d, (void *)lacl);
	free ((void *)lacl);
	return error ? -1 : 0;
}

int		acl_set_fd (const int fd, const acl_type_t type_d,
			    const acl_t acl_d)
{
	return acl_set (LINUX_ACL_SET_FD, (void *)fd, type_d, acl_d);
}

int		acl_set_file (const char *path_p, const acl_type_t type_d,
			      const acl_t acl_d)
{
	return acl_set (LINUX_ACL_SET_FILE, (void *)path_p, type_d, acl_d);
}
