/*
 * 1999-02-04  Hauke Steenbock <steenbo@informatik.uni-bremen.de>
 * remove_entries: uidp2 and gidp2 get the right qualifier
 * add_entries: uidp2 and gidp2 get the right qualifier
 */

#include <stdio.h>
#include <string.h>
#include <sys/acl.h>
#include <sys/param.h>
#include <errno.h>
#include <sys/stat.h>
#include <unistd.h>

#include "util.h"

#define	CHUNK		1024

extern const char *__progname;

static const char *program_name;

static
volatile void	usage (void)
{
	fprintf (stderr, "Usage: %s [-bdkn] [-m entries] [-M file1]"
					 " [-x entries] [-X file2] [file...]\n", program_name);
	exit (1);
}

static
int		remove_non_base (acl_t acl)
{
	acl_entry_t	acle1;
	acl_entry_t	acle2;
	acl_tag_t	tag;

#ifdef	DEBUG
	printf ("remove_non_base called on ");
	print_acl (stdout, acl);
	printf ("\n");
#endif
	if (acl_first_entry (acl) == -1)
		return -1;
	for (acle1 = acl_get_entry (acl); acle1; acle1 = acle2) {
		acle2 = acl_get_entry (acl);
		if (acl_get_tag_type (acle1, &tag) == -1)
			return -1;
		if (tag == ACL_USER || tag == ACL_GROUP || tag == ACL_MASK)
			if (acl_delete_entry (acl, acle1) == -1)
				return -1;
	}
#ifdef	DEBUG
	printf ("remove_non_base changed ACL to ");
	print_acl (stdout, acl);
	printf ("\n");
#endif
	return 0;
}

static
int		remove_entries (acl_t acl, acl_t entries, int *foundmask)
{
	acl_entry_t	acle1;
	acl_entry_t	acle2;
	acl_entry_t	acle3;
	acl_tag_t	tag1;
	acl_tag_t	tag2;
	uid_t		*uidp1;
	uid_t		*uidp2;
	gid_t		*gidp1;
	gid_t		*gidp2;

#ifdef	DEBUG
	printf ("remove_entries called to remove ");
	print_acl (stdout, entries);
	printf (" from ");
	print_acl (stdout, acl);
	printf ("\n");
#endif
	*foundmask = 0;
	if (acl_first_entry (entries) == -1)
		return -1;
	for (acle1 = acl_get_entry (entries); acle1;
			 acle1 = acl_get_entry (entries)) {
		if (acl_get_tag_type (acle1, &tag1) == -1)
			return -1;
		uidp1 = NULL;
		gidp1 = NULL;
		switch (tag1) {
		case ACL_MASK:
			*foundmask = 1;
			break;
		case ACL_USER:
			uidp1 = (uid_t *)acl_get_qualifier (acle1);
			if (!uidp1)
				return -1;
			break;
		case ACL_GROUP:
			gidp1 = (gid_t *)acl_get_qualifier (acle1);
			if (!gidp1)
				return -1;
			break;
		default:
			break;
		}
		if (acl_first_entry (acl) == -1)
			return -1;
		for (acle2 = acl_get_entry (acl); acle2; acle2 = acle3) {
			acle3 = acl_get_entry (acl);
			if (acl_get_tag_type (acle2, &tag2) == -1)
				return -1;
			if (tag1 != tag2)
				continue;
			uidp2 = NULL;
			gidp2 = NULL;
			switch (tag2) {
			case ACL_USER:
				uidp2 = (uid_t *)acl_get_qualifier (acle2);
				if (!uidp2)
					return -1;
				if (*uidp1 == *uidp2)
					if (acl_delete_entry (acl, acle2) == -1)
						return -1;
				if (acl_free_qualifier (uidp2,
																tag2) == -1)
					return -1;
				break;
			case ACL_GROUP:
				gidp2 = (gid_t *)acl_get_qualifier (acle2);
				if (!gidp2)
					return -1;
				if (*gidp1 == *gidp2)
					if (acl_delete_entry (acl, acle2) == -1)
						return -1;
				if (acl_free_qualifier (gidp2,
																tag2) == -1)
					return -1;
				break;
			default:
				if (acl_delete_entry (acl, acle2) == -1)
					return -1;
				break;
			}
		}
		if (uidp1)
			if (acl_free_qualifier (uidp1, tag1) == -1)
				return -1;
		if (gidp1)
			if (acl_free_qualifier (gidp1, tag1) == -1)
				return -1;
	}
#ifdef	DEBUG
	printf ("remove_entries changed ACL to ");
	print_acl (stdout, acl);
	printf ("\n");
#endif
	return 0;
}

static
int		update_entry (acl_entry_t acle1, acl_entry_t acle2,
										acl_mode_t mode)
{
	acl_permset_t	perms1;
	acl_permset_t	perms2;

	if (acl_get_permset (acle1, &perms1) == -1 ||
			acl_get_permset (acle2, &perms2) == -1)
		return -1;
	switch (mode) {
	case ACL_ABSOLUTE_ENTRY:
		if (acl_clear_perms (perms2) == -1)
			return -1;
		/* Fall through */
	case ACL_ADD_ENTRY:
		if (acl_add_perm (perms2, *perms1) == -1)
			return -1;
		break;
	case ACL_REMOVE_ENTRY:
		if (acl_delete_perm (perms2, *perms1) == -1)
			return -1;
		break;
	default:											/* should not happen */
		errno = EINVAL;
		return -1;
	}
	if (acl_set_permset (acle2, perms2) == -1)
		return -1;
	return 0;
}

static
int		add_entry (acl_t *acl, acl_entry_t src_acle)
{
	acl_entry_t	acle;

	if(acl_create_entry (acl, &acle) != 0) 
		return -1;
	return acl_copy_entry (acle, src_acle);
}

static
int	add_empty_entry (acl_t *acl, acl_entry_t src_acle)
{
  acl_permset_t	perms, org;
  if (acl_get_permset (src_acle, &perms) == -1)
    return -1;
  if (acl_get_permset (src_acle, &org) == -1)
    return -1;
  if (acl_clear_perms (perms) == -1)
    return -1;
  if (acl_set_permset (src_acle, perms) == -1)
    return -1;
  if (add_entry (acl, src_acle) == -1)
    return -1;
  return acl_set_permset (src_acle, org);
}

static
int		update_entries (acl_t acl, acl_t entries, int *foundmask)
{
	acl_entry_t	e_from;
	acl_entry_t	e_to;
	acl_entry_t	tmp;
	acl_tag_t	tag_from;
	acl_tag_t	tag_to;
	acl_mode_t	mode;
	uid_t		*uid_from;
	uid_t		*uid_to;
	gid_t		*gid_from;
	gid_t		*gid_to;
	int		done;

	*foundmask = 0;
	if (acl_first_entry (entries) == -1)
		return -1;
	while ((e_from = acl_get_entry (entries))) {
		uid_from = NULL;
		gid_from = NULL;
		if (acl_get_tag_type (e_from, &tag_from) == -1)
			return -1;
		switch (tag_from) {
		case ACL_MASK:
			*foundmask = 1;
			break;
		case ACL_USER:
			uid_from = (uid_t *)acl_get_qualifier (e_from);
			if (!uid_from)
				return -1;
			break;
		case ACL_GROUP:
			gid_from = (gid_t *)acl_get_qualifier (e_from);
			if (!gid_from)
				return -1;
			break;
		default:
			break;
		}
		mode = acl_get_mode (e_from);
		if (mode == -1)
			return -1;
		if (acl_first_entry (acl) == -1)
			return -1;
		done = 0;
		for (e_to = acl_get_entry (acl); e_to; e_to = tmp) {
			tmp = acl_get_entry (acl);
			if (acl_get_tag_type (e_to, &tag_to) == -1)
				return -1;
			if (tag_from != tag_to)
				continue;
			uid_to = NULL;
			gid_to = NULL;
			switch (tag_to) {
			case ACL_USER:
				uid_to = (uid_t *)acl_get_qualifier (e_to);
				if (!uid_to)
					return -1;
				if (*uid_from == *uid_to) {
					if (update_entry (e_from, e_to, mode) == -1)
						return -1;
					done = 1;
				}
				if (acl_free_qualifier (uid_to, tag_to) == -1)
					return -1;
				break;
			case ACL_GROUP:
				gid_to = (gid_t *)acl_get_qualifier (e_to);
				if (!gid_to)
					return -1;
				if (*gid_from == *gid_to) {
					if (update_entry (e_from, e_to, mode) == -1)
						return -1;
					done = 1;
				}
				if (acl_free_qualifier (gid_to, tag_to) == -1)
					return -1;
				break;
			default:
				if (update_entry (e_from, e_to, mode) == -1)
					return -1;
				done = 1;
				break;
			}
		}
		if (!done) {
			if (mode == ACL_REMOVE_ENTRY){
        if(add_empty_entry(&acl, e_from) == -1)
          return -1;
			} else if (add_entry (&acl, e_from) == -1)
				return -1;
		}
		if (uid_from) {
			if (acl_free_qualifier (uid_from, tag_from) == -1)
				return -1;
		}
		if (gid_from) {
			if (acl_free_qualifier (gid_from, tag_from) == -1)
				return -1;
		}
	}
	return 0;
}

static
acl_t		acl_from_file (const char *file)
{
	char		*text = NULL;
	FILE		*f;
	acl_t		acl;
	int		closeit = 0, size = 0, pos = 0;

	if (strcmp(file,"-") == 0) {
		f = stdin;
	} else {
		f = fopen (file, "r");
		closeit = 1;
	}
	if (!f)
		return NULL;
	while (!feof (f)) {
		char *tmp = text;
		text = realloc(text, size + CHUNK + 1);
		if(text == NULL){
			free(tmp);
			if(closeit)
				fclose(f);
			return NULL;
		}
		pos += fread(&text[pos], 1, CHUNK, f);
	}
	text[pos] = 0;
	acl = acl_from_text (text);
	free (text);
	if(closeit)
		fclose(f);
	return acl;
}

int 
main (int argc, char *argv[])
{
	acl_t		access_acl, default_acl, acl, modify;
	int		i;
	char	c;
	struct stat stat_buffer;
	char from_stdin[MAXPATHLEN];
	char *current;
	int isdir;
	int stdinused = 0;
	int access_modified, default_modified;
	int foundmask, missing_mask;
	int   optn = 0, optd = 0;

	program_name = __progname;

	/* test for -d option */
	while ((c = getopt (argc, argv, "bdknm:M:x:X:")) != EOF) {
		switch(c){
		case 'd':
			optd = 1;
			break;
		case 'n':
			optn = 1;
			break;
		}
	}
	for(i = optind; i < argc; i++){
	read_from_stdin:
		if(strcmp(argv[i], "-") == 0){
			if(stdinused)
				goto error_stdin;
			stdinused = 1;
			if(fgets (from_stdin, sizeof (from_stdin), stdin) == NULL)
				continue;
			strtok(from_stdin,"\n");
			current = from_stdin;
		} else {
			current = argv[i];
		}
		if(stat(current, &stat_buffer) < 0){
			fprintf(stderr, strerror (errno));
			goto next_one;
		}
		isdir = S_ISDIR(stat_buffer.st_mode);
		if(!optd)
			access_acl = acl_get_file (current, ACL_TYPE_ACCESS);
		if(isdir)
			default_acl = acl_get_file (current, ACL_TYPE_DEFAULT);
		access_modified = 0;
		default_modified = 0;
		missing_mask = 0;
		optind = 0;
		while ((c = getopt (argc, argv, "bdknm:M:x:X:")) != EOF) {
			/* perhaps invalidated by option */
			acl = optd ? default_acl : access_acl;
			switch (c)   {
			case 'd':
			case 'n':
				break;
			case 'b':
				if (remove_non_base(acl) == -1) {
					perror_with_acl (acl, "remove_non_base");
					exit(2);
				}
				break;
			case 'k':
				if(!isdir){
					fprintf(stderr, "-k option only allowed with directory");
					goto next_one;
				}
				acl_free(default_acl);
				default_acl = acl_init (sizeof (struct acl));
				default_modified = 1;
				break;
			case 'm':
			case 'x':
				modify = acl_from_text (optarg);
				if (!modify) {
					perror_with_arg ("acl_from_text", optarg);
					exit (2);
				}
				break;
			case 'M':
			case 'X':
				if(strcmp(optarg,"-") == 0) {
					if(stdinused)
						goto error_stdin;
					stdinused = 1;
				}
				modify = acl_from_file (optarg);
				if (!modify) {
					perror_with_arg ("acl_from_file", optarg);
					exit (2);
				}
				break;
			default:
				usage ();
			}
			switch (c)   {
			case 'm':
			case 'M':
				if(update_entries (acl, modify, &foundmask) == -1){
					perror_with_acl (acl, "update_entries");
					exit(2);
				}
				acl_free(modify);
				break;
			case 'x':
			case 'X':
				if(remove_entries (acl, modify, &foundmask) == -1){
					perror_with_acl (acl, "update_entries");
					exit(2);
				}
				acl_free(modify);
				break;
			}
			switch (c)   {
			case 'm':
			case 'M':
			case 'x':
			case 'X':
				if(!foundmask){
					missing_mask = 1;
				}
				/* fall thru */
			case 'b':
				if(optd)
					default_modified = 1;
				else
					access_modified = 1;
				break;
			}
		}
		if(default_modified){
			if(default_acl->acl_used_acle > 0){
				if(optd && missing_mask && !optn){
					if (acl_calc_mask(&default_acl) == -1){
						perror_with_acl (default_acl, "acl_calc_mask");
						exit(1);
					}
				}
				if (acl_valid (default_acl, NULL) == -1){
					perror_with_acl (default_acl, "acl_valid");
					exit(2);
				}
			}
		}
		if(access_modified){
			if(!optd && missing_mask && !optn){
				if (acl_calc_mask(&access_acl) == -1){
					perror_with_acl (access_acl, "acl_calc_mask");
					exit(1);
				}
			}
			if (acl_valid (access_acl, NULL) == -1){
				perror_with_acl (access_acl, "acl_valid");
				exit(2);
			}
		}
		if(default_modified){
			if(default_acl->acl_used_acle == 0){
				if(acl_delete_def_file(current) == -1){
					perror_with_acl_and_arg (default_acl, "acl_delete_def_file", current);
					exit(1);
				}
			} else {
				if(acl_set_file (current, ACL_TYPE_DEFAULT, default_acl) == -1){
					perror_with_acl_and_arg (default_acl, "acl_set_file", current);
					exit(1);
				}
			}
		}
		if(access_modified){
			if(acl_set_file (current, ACL_TYPE_ACCESS, access_acl) == -1){
				perror_with_acl_and_arg (access_acl, "acl_set_file", current);
				exit(1);
			}
		}
	next_one:
		acl_free(default_acl);
		acl_free(access_acl);
		if(current == from_stdin)
			goto read_from_stdin;
	}
	exit(0);
 error_stdin:
	fprintf (stderr, "%s: only one usage of stdin (-) is allowed.\n",
					 program_name);
	exit (2);
}
